/*
 * SPDX-FileCopyrightText: 2024-2024 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#ifndef _FCITX_RIMEFACTORY_H_
#define _FCITX_RIMEFACTORY_H_

#include <fcitx/addonfactory.h>

namespace fcitx::rime {

class RimeEngineFactory : public AddonFactory {
public:
    AddonInstance *create(AddonManager *manager) override;
};

} // namespace fcitx::rime

#endif
